#include "ddevice.h"
#include "input/codes.h"

#include "mmgr.h"

DInputDevice::DInputDevice(IDirectInputDevice8* device, DInputManager* manager)
{
	Listener = 0;
	Data = 0;
	ValueData = 0;
	StateData = 0;

	Device = device;
	Manager = manager;

	// record device name
	DIDEVICEINSTANCE dinstance;
	dinstance.dwSize = sizeof(dinstance);
	Device->GetDeviceInfo( &dinstance);

	Type = strlwr(dinstance.tszInstanceName);

	// record number of values and states
	DIDEVCAPS  caps; 
 
	caps.dwSize = sizeof(DIDEVCAPS); 
	Device->GetCapabilities(&caps); 

	Axes = caps.dwAxes;
	POVs = caps.dwPOVs;
	
	Values = Axes + POVs;
	States= caps.dwButtons;
	
	// set up the cooperative level
	Device->SetCooperativeLevel(GetActiveWindow(),
		DISCL_NONEXCLUSIVE | DISCL_BACKGROUND | DISCL_NOWINKEY);

	// now enumerate all the device pieces and set the 
	// data mode
	SetFormat();
	// set up the return data structure
	ReturnData.Set( this, ValueData, Values, StateData, States);

	// get first set of values
	SetShortCuts();
	Update();
}

DInputDevice::~DInputDevice()
{
	if (Device)
		Device->Release();
	if (Data)
		delete [] Data;
}

gs_InputData& DInputDevice::GetInput()
{
	return ReturnData;
}

void DInputDevice::Update()	
{
	// this updates the input
	// get data in a buffered format, then send to all the 
	// classes that want buffered data, and update the state memory
	
	int error = Device->Poll();

	if (error == DIERR_NOTACQUIRED)
	{
		Device->Acquire();
		error = Device->Poll();
		if (error == DIERR_NOTACQUIRED)
			return;
	}

	// get current data
	Device->GetDeviceState(DataSize, Data);

	// if any objects want to recieve events, get buffered data

	if (Listener)
	{
		DIDEVICEOBJECTDATA data[BUFFERSIZE];
		unsigned long l = BUFFERSIZE;
		Device->GetDeviceData( sizeof(DIDEVICEOBJECTDATA), 
								data,
								&l,
								0);

		// REMEMBER TO CHECK if error == buffer input lost

		// iterate through all the data in the list
		for (int x = 0; x < l; x++)
		{
			InputState tosend;
			tosend.Parent = this;
			tosend.Time = data[x].dwTimeStamp;

			// data[x].dwOfs
			if (data[x].dwOfs < Values*4)
				tosend.Index = data[x].dwOfs / 4;
			else
				tosend.Index = data[x].dwOfs - (3*Values);
		
			tosend.Data = (int)data[x].dwData;			// actual value
	
			// then send it to all that have requested it
			Listener->Message(&tosend);
		}
	}
	
}

void DInputDevice::SetAxisMode( AxisStates state )
{
	int error;

	Device->Unacquire();
	// set up the buffer size
	DIPROPDWORD  dipdw; 
	dipdw.diph.dwSize = sizeof(DIPROPDWORD); 
	dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
	dipdw.diph.dwObj = 0; 
	dipdw.diph.dwHow = DIPH_DEVICE; 

	if (state == AxisRelative)
		dipdw.dwData = DIPROPAXISMODE_REL;
	else
		dipdw.dwData = DIPROPAXISMODE_ABS;

	error = Device->SetProperty(DIPROP_AXISMODE , &dipdw.diph);
}

void DInputDevice::SetFormat()
{
	int offset, x;
	offset = 0;
	DIDATAFORMAT format;
	format.dwSize = sizeof(format);
	format.dwObjSize = sizeof( DIOBJECTDATAFORMAT );

	if ( Type.find( "mouse") != -1)
		format.dwFlags = DIDF_RELAXIS;
	else
		format.dwFlags = DIDF_ABSAXIS;

	DIOBJECTDATAFORMAT* array = new DIOBJECTDATAFORMAT[Values+States];

	DIOBJECTDATAFORMAT toadd; 
	toadd.pguid = 0; // wildcard 
    toadd.dwFlags = 0; // ignore aspect 
	
	for (x = 0; x < Axes; x++)
	{
		toadd.dwOfs = offset; // byte offset
		toadd.dwType = DIDFT_AXIS | DIDFT_ANYINSTANCE; // any axis
		offset+=4;
		array[x] = toadd;
	}

	for (x = 0; x < POVs; x++)
	{
		toadd.dwOfs = offset; // byte offset
		toadd.dwType = DIDFT_POV | DIDFT_ANYINSTANCE; // any pov switch
		offset+=4;
		array[x+Axes] = toadd;
	}

	for (x = 0; x < States; x++)
	{
		toadd.dwOfs = offset; // byte offset
		toadd.dwType = DIDFT_BUTTON | DIDFT_ANYINSTANCE; // any button 
		offset+=1;
		array[x+Axes+POVs] = toadd;
	}

	DataSize = offset + (4-(offset%4));

	format.dwDataSize = DataSize; // Size of a data packet returned by the device, in bytes. This value must be a multiple of 4 and must exceed the largest offset value for an object's data within the data packet. 
	format.dwNumObjs = Values + States; // Number of objects in the rgodf array. 
	format.rgodf = array; // Address to an array 

	// special keyboard only format, used for proper indexing
	if (Type.find("keyboard") != -1)
	{
		format = c_dfDIKeyboard;
		DataSize = 256;
		States = 256;
	}



	int error = Device->SetDataFormat( &format);

	delete [] array;

	// now set up the data arrays for getting input
	Data = new char[DataSize];
	memset(Data,0, DataSize);
	ValueData = (int*)&Data[0];
	StateData = (unsigned char*)&Data[Values*4];

}

void DInputDevice::SetShortCuts()
{
	if (GetType().find("mouse") != -1)
	{
		MOUSEX = 0;
		MOUSEY = 1;

		if (Axes > 2)
			MOUSEZ = 2;
		else
			MOUSEZ = -1;
		
		MOUSEBUTTON0 = Values + 0;
		MOUSEBUTTON1 = Values + 1;
		
		if (States > 2)
			MOUSEBUTTON2 = Values + 2;
		else
			MOUSEBUTTON2 = -1;
	}
	
	if (GetType().find("joystick") != -1)
	{
		JOYSTICKX = 0;
		JOYSTICKY = 1;

		if (Axes > 2)
			JOYSTICKZ = 2;
		else
			JOYSTICKZ = -1;
		
		JOYSTICKBUTTON0 = Values + 0;
		JOYSTICKBUTTON1 = Values + 1;
		
		if (States > 2)
			JOYSTICKBUTTON2 = Values + 2;
		else
			JOYSTICKBUTTON2 = -1;

		if (States > 3)
			JOYSTICKBUTTON3 = Values + 3;
		else
			JOYSTICKBUTTON3 = -1;

	}

}


BOOL CALLBACK DIEnumDeviceObjectsCallback( 
                      LPCDIDEVICEOBJECTINSTANCE lpddoi,
                      LPVOID pvRef) 
{ 
	DInputDevice* d = (DInputDevice*)pvRef;

	if (d->CheckName( (char*)lpddoi->tszName ) != 0)
		return DIENUM_STOP;

	return DIENUM_CONTINUE; 
} 

int DInputDevice::CheckName( char* name)
{
	if (CheckMode == 1)
	{
		if (toCheck.find( strlwr(name) ) != -1)
		{
			Found = Enumerated;
			return 1;
		}
	}
	else
	{
		if (Enumerated == Found)
		{
			toCheck = name;
			return 1;
		}
	}

	Enumerated++;
	return 0;
}

int DInputDevice::GetIndex( string name )
{
	Enumerated = 0;
	Found = -1;
	toCheck = name.c_str();	// force to copy string, otherwise uses pointer to same resource and tries to delete memory across dll
	CheckMode = 1;
	// this is probably slow, but it works

	// enumerate all the objects
	// check readable name
	// if readable name == name
	// then return index
	
	// if none work, return -1
	Device->EnumObjects(DIEnumDeviceObjectsCallback, this, DIDFT_AXIS); 
	Device->EnumObjects(DIEnumDeviceObjectsCallback, this, DIDFT_POV); 
	Device->EnumObjects(DIEnumDeviceObjectsCallback, this, DIDFT_BUTTON); 

	return Found;
}

string DInputDevice::GetName( int index)
{
	toCheck = "";
	CheckMode = 2;
	Enumerated = 0;
	Found = index;
	return toCheck;	
}

int DInputDevice::AddNotify( gs_InputListener* b)
{
	// if there are no listeners, the input buffer
	// is not needed, if there are, make sure there is a buffer
	SetInputBuffer(BUFFERSIZE);

	Listener = b;
	return 1;
}

int DInputDevice::RemoveNotify( gs_InputListener* b)
{
	Listener = 0;

	// if there are no listeners, the input buffer
	// is not needed
	SetInputBuffer(0);

	return 1;
}

void DInputDevice::SetInputBuffer(int size)
{
	Device->Unacquire();
	// set up the buffer size
	DIPROPDWORD  dipdw; 
	dipdw.diph.dwSize = sizeof(DIPROPDWORD); 
	dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
	dipdw.diph.dwObj = 0; 
	dipdw.diph.dwHow = DIPH_DEVICE; 
	dipdw.dwData = BUFFERSIZE; // buffer size
	int error = Device->SetProperty(DIPROP_BUFFERSIZE, &dipdw.diph); 
}